import os
import dill
import numpy as np
from math import pi, exp, atan, sqrt, acos


def angle(x, y):
	"""calculates the angle of (x,y) with respect to (0,0)
	
	:param x: x choordinate
	:param y: y choordinate
	:returns: the angle"""
	at = atan(y/x)
	if(x < 0): return at+pi
	elif(y < 0): return at+2*pi
	return at


def angle_between_vectors(v, w):
	"""calculates the angle between two vectors (any dimensional)
	
	:param v: vector 1
	:param w: vector 2
	:returns: the angle"""
	scalar = sum(v[i]*w[i] for i in range(len(v)))
	if(scalar/(L2_norm(v)*L2_norm(w)) > 1): return 0
	if(scalar/(L2_norm(v)*L2_norm(w)) < -1): return pi
	return acos(scalar/(L2_norm(v)*L2_norm(w)))


def L1_norm(seq):
	"""calculates the L1 norm of a sequence or vector
	
	:param seq: the sequence or vector
	:returns: the L1 norm"""
	norm = 0
	for i in range(len(seq)):
		norm += abs(seq[i])
	return norm


def L2_norm(seq):
	"""calculates the L2 norm of a sequence or vector
	
	:param seq: the sequence or vector
	:returns: the L2 norm"""
	norm = 0
	for i in range(len(seq)):
		norm += seq[i]**2
	return sqrt(norm)


def Linf_norm(seq):
	"""calculates the Linf norm of a sequence or vector
	
	:param seq: the sequence or vector
	:returns: the Linf norm"""
	largest = 0
	for i in range(len(seq)):
		if(abs(seq[i]) > largest): largest = abs(seq[i])
	return largest


def linear_interpolation(listx, listy, argument):
	"""calculates the linear interpolation of [listx,listy] at argument
	
	:param listx: x choordinates (should be ordered in ascending order)
	:param listy: y choordinates
	:param argument: where to evaluate the linear interpolation
	:returns: value of the linear interpolation at argument"""
	if(argument in listx):
		return listy[listx.index(argument)]
	if(argument < listx[0]):
		return listy[0] + (listy[0]-listy[1])/(listx[1]-listx[0])*(listx[0]-argument)
	if(argument > listx[-1]):
		return listy[-1] + (listy[-1]-listy[-2])/(listx[-1]-listx[-2])*(argument-listx[-1])
	index = 0
	while((listx[index] < argument and listx[index+1] > argument) != 1): index += 1
	return listy[index] + (listy[index+1]-listy[index])/(listx[index+1]-listx[index])*(argument-listx[index])


def gaussian(x, std=1):
	"""returns a Gaussian distribution
	
	:param x: variable
	:param std: standard deviation (default 1)
	:returns: Gaussian PDF"""
	return 1/(std*sqrt(2*pi))*exp(-x**2/(2*std**2))


def gaussian_der(x, std=1):
	"""returns a Gaussian derivative distrbution (x*PDF_normal)
	
	:param x: variable
	:param std: the standard deviation
	:returns: positive half of the Gaussian derivative PDF"""
	if(x < 0): return 0
	return (2-pi/2)/std**2*x*exp(-(2-pi/2)*x**2/(2*std**2))


def kernel_convolve(seq, kernel, kernel_width):
	"""convolves a sequence with a kernel
	
	:param seq: sequence to be convolved
	:param kernel: the kernel, a probability distribuion fucntion (does not have to be normalized)
	:param kernel_width: the width of the kernel (std for Gaussian)
	:returns: convolved sequence"""
	ker = np.array([kernel(i/kernel_width) for i in range(-round(3*kernel_width), round(3*kernel_width))])
	ker = ker/sum(ker)
	return np.convolve(seq, ker ,mode='same')


def save_object_dill(obj, filename, save_path="objects_save"):
	"""saves an object with dill (also makes the 'objects_save/' directory if it wasn't there)
	
	:param obj: object to save
	:param filename: filename without any extension
	:param save_path: path where to save object, staring from the directory of the simulation (default "objects_save")"""
	# directory
	if not os.path.exists(save_path):
		os.makedirs(save_path)
	# pickling
	f = open(save_path+"/"+filename+'.pickle', 'wb')	
	dill.dump(obj, f)
	f.close()
	
	
def load_object_dill(filename, load_path="objects_save"):
	"""loads an object with dill (from 'objects_save/' directory)
	
	:param filename: filename without any extension
	:param load_path: path where to load object from, staring from the directory of the simulation (default "objects_save")
	:returns: saved object"""
	# unpickling
	f = open(load_path+"/"+filename+'.pickle', 'rb')	
	obj = dill.load(f)
	f.close()
	return obj




import numpy as np
from math import pi, log, exp, sqrt, floor, sin, cos, atan
import random

from oscillator_snap.oscillator_auxiliaries import *

	
def one_step_integrator(state, ders, I, dt):
	"""RK4 integrates state with derivative and input for one step of dt
	
	:param state: state of the variables
	:param ders: derivative functions
	:param I: input
	:param dt: time step
	:return: state after one integration step"""
	D = len(state)
	# 1
	k1 = [ders[i](state,I) for i in range(D)]
	# 2
	state2 = [state[i]+k1[i]*dt/2.0 for i in range(D)]
	k2 = [ders[i](state2,I) for i in range(D)]
	# 3
	state3 = [state[i]+k2[i]*dt/2.0 for i in range(D)] 
	k3 = [ders[i](state3,I) for i in range(D)]
	# 4
	state4 = [state[i]+k3[i]*dt for i in range(D)] 
	k4 = [ders[i](state4,I) for i in range(D)]
	# put togeather
	statef = [state[i] + (k1[i]+2*k2[i]+2*k3[i]+k4[i])/6.0*dt for i in range(D)]
	return statef


def oscillator_period(ders, inp, warmup_time=1000.0, thr=0.0, dt=0.01):
	"""calculates the natural period of the oscillator
	
	:param ders: a list of state variable derivatives
	:param inp: the input to the system
	:param warmup_time: the time for relaxing to the stable orbit (default 1000)
	:param thr: threshold for determining period (default 0.0)
	:param dt: time step (default 0.01)
	:returns: natural period"""
	# initial conditions
	state = [0.0 for i in range(len(ders))]
	state[0] = 1.0
	# warmup
	for i in range(round(warmup_time/dt)):
		state = one_step_integrator(state, ders, inp, dt)
	# integration up to x = thr
	xh = state[0]
	while((state[0] > thr and xh < thr) == False):
		xh = state[0]
		state = one_step_integrator(state, ders, inp, dt)
	# Henon trick
	dt_beggining = 1.0/ders[0](state,inp)*(state[0]-thr)
	# spoil condition and go again to x = 0 (still counting time)
	xh = state[0]
	time = 0
	while((state[0] > thr and xh < thr) == False):
		xh = state[0]
		state = one_step_integrator(state, ders, inp, dt)
		time = time + dt
	# another Henon trick
	dt_end = 1.0/ders[0](state,inp)*(state[0]-thr)
	return time + dt_beggining - dt_end


def oscillator_average_period(ders, inp, warmup_time=1000.0, avg_time=5000.0, thr=0.0, dt=0.01):
	"""calculates the average period of the oscillator (for chaotic or quasi-periodic)
	
	:param ders: a list of state variable derivatives
	:param inp: the input to the system
	:param warmup_time: the time for relaxing to the stable orbit (default 1000)
	:parma avg_time: averaging time (default 5000)
	:param thr: threshold for determining period (default 0.0)
	:param dt: time step (default 0.01)
	:returns: average period"""
	# initial conditions
	state = [0.0 for i in range(len(ders))]
	state[0] = 1.0
	# warmup
	for i in range(round(warmup_time/dt)):
		state = one_step_integrator(state, ders, inp, dt)
	# integration up to x = thr
	xh = state[0]
	while((state[0] > thr and xh < thr) == False):
		xh = state[0]
		state = one_step_integrator(state, ders, inp, dt)
	# start counting periods and measuring time (taking into account the amount over x = 0)
	periods = 0
	time = (state[0]-thr)/ders[0](state,inp)
	# run for avg_time
	for i in range(round(avg_time/dt)):
		xh = state[0]
		state = one_step_integrator(state, ders, inp, dt)
		time = time + dt
		if(state[0] > 0 and xh < 0): periods = periods+1
	# spoil condition and go again to x = 0 (still counting time)
	xh = state[0]
	while((state[0] > thr and xh < thr) == False):
		xh = state[0]
		state = one_step_integrator(state, ders, inp, dt)
		time = time + dt
	# correct for going over x = thr
	time = time - (state[0]-thr)/ders[0](state,inp)
	return time/(periods+1)


def oscillator_PRC(ders, inp=0.0, warmup_time=1000.0, stimulation=1.0, period_counts=5, dph=0.1, thr=0.0, dt=0.01):
	"""calculates the PRC
	
	:param ders: a list of state variable derivatives
	:param inp: the offset input to the system (default 0.0)
	:param warmup_time: the time for relaxing to the stable orbit (default 1000)
	:param stimulation: strength of the stimulation (default 1.0)
	:param period_counts: how many periods to wait for evaluating the asymptotic phase shift (default 5)
	:param dph: phase resolution (default 0.1)
	:param thr: threshold for determining period (default 0.0)
	:param dt: time step (default 0.01)
	:returns: average period"""
	period = oscillator_period(ders, inp)
	PRC = [[dph*i for i in range(floor(2*pi/dph))],[0 for i in range(floor(2*pi/dph))]] # PRC list
	# initial conditions
	state = [0.0 for i in range(len(ders))]
	state[0] = 1.0
	# warmup
	for i in range(round(warmup_time/dt)):
		state = one_step_integrator(state, ders, inp, dt)
	# stimulating phases
	for ph in [dph*i for i in range(floor(2*pi/dph))]:
		# integration up to x = thr
		xh = state[0]
		while((state[0] > thr and xh < thr) == False):
			xh = state[0]
			state = one_step_integrator(state, ders, inp, dt)
		# Henon trick
		dt_beggining = 1.0/ders[0](state,inp)*(state[0]-thr)
		# spoil condition and go to ph (counting time)
		xh = state[0]
		time = dt_beggining
		while(time < ph/(2*pi)*period):
			xh = state[0]
			state = one_step_integrator(state, ders, inp, dt)
			time = time + dt
		# stimulate 
		xh = state[0]
		state = one_step_integrator(state, ders, inp+stimulation, dt) # notice +stimulation
		time = time + dt
		#integrate for some periods
		for p in range(period_counts):
			xh = state[0] # spoil
			while((state[0] > thr and xh < thr) == False):
				xh = state[0]
				state = one_step_integrator(state, ders, inp, dt)
				time = time + dt
		# another Henon trick
		dt_end = 1.0/ders[0](state,inp)*(state[0]-thr)
		phase_shift = 2*pi*(time-dt_end - period_counts*period)/period
		PRC[1][round(ph/dph)] = phase_shift/stimulation
	return PRC


def oscillator_lyapunov(ders, inp=0.0, warmup_time=1000.0, delta=0.00001, tau=1.0, unconsidered_trials=100, trials=1000, dt=0.01):
	"""calculates the largest Lyapunov exponent
	
	:param ders: a list of state variable derivatives
	:param inp: the offset input to the system (default 0.0)
	:param warmup_time: the time for relaxing to the stable orbit (default 1000)
	:param delta: the amplitude of the initial trajectory shift (default 0.00001)
	:param tau: the time for which the two trajectories are left to deviate (default 1.0)
	:param unconsidered_trials: the number of trials that are not to be taken into account when averaging but allow the largest component of the Lyapunov vector to become dominant (default 100)
	:param trials: the number of trials over which we average to get an estimation of the Lyapunov exponent (default 1000)
	:param dt: time step (default 0.01)
	:returns: largest Lyapunov exponent"""
	# initial conditions
	state = [0.0 for i in range(len(ders))]
	state[0] = 1.0
	# warmup
	for t in range(round(warmup_time/dt)):
		state = one_step_integrator(state, ders, inp, dt)
	# other trajectory init
	state2 = state[:]
	state2[0] += delta # deviation
	# uncosidered trials
	for i in range(unconsidered_trials):
		for t in range(round(tau/dt)):
			state = one_step_integrator(state, ders, inp, dt)
			state2 = one_step_integrator(state2, ders, inp, dt)
		diff = np.array(state)-np.array(state2) # difference
		diff = delta*diff/L2_norm(diff) # normalize (to delta)
		state2 = [state[0]+diff[0], state[1]+diff[1], state[2]+diff[2]]# readjust state2
	# cosidered trials
	lyapunov = 0.0
	for i in range(trials):
		for t in range(round(tau/dt)):
			state = one_step_integrator(state, ders, inp, dt)
			state2 = one_step_integrator(state2, ders, inp, dt)
		diff = np.array(state)-np.array(state2) # difference
		lyapunov += log(L2_norm(diff)/delta) # averaging!
		diff = delta*diff/L2_norm(diff) # normalize (to delta)
		state2 = [state[0]+diff[0], state[1]+diff[1], state[2]+diff[2]]# readjust state2
	return lyapunov/tau/trials


def oscillator_bifurcation(ders, inp_min, inp_max, d_inp, warmup_time=1000.0, short_warmup=200.0, time_window=500, dt=0.01):
	"""calculates the bifurcation diagram with respect to the input
	
	:param ders: a list of state variable derivatives
	:param inp_min: the min input
	:param inp_max: the max input
	:param d_inp: input resolution
	:param warmup_time: the time for relaxing to the stable orbit (default 1000)
	:param short_warmup: the time for relaxing between different inputs (default 200)
	:param time_window: time window (default 500)
	:param dt: time step (default 0.01)
	:returns: bifurcation diagram list"""
	print("Calculating biffurcation diagram...")
	bif = [[],[]] # bifurcation list
	# initial conditions
	state = [0.0 for i in range(len(ders))]
	state[0] = 1.0
	# warmup
	for i in range(round(warmup_time/dt)):
		state = one_step_integrator(state, ders, inp_min, dt)
	# inp loop
	for inp in [inp_min+d_inp*i for i in range(floor((inp_max-inp_min)/d_inp))]:
		print("\tb = ", "%.3f" % inp, "/", inp_max)
		xhhhh = state[0] # pre-pre-pre-previous value
		xhhh = state[0] # pre-pre-previous value
		xhh = state[0] # pre-previous value
		xh = state[0] # previous value
		# warmup
		for i in range(round(short_warmup/dt)):
			state = one_step_integrator(state, ders, inp, dt)
		for t in range(round(time_window/dt)):
			xhhhh = xhhh
			xhhh = xhh
			xhh = xh
			xh = state[0]
			state = one_step_integrator(state, ders, inp, dt)
			if(xhhhh < xhhh and xhhh < xhh and xhh > xh and xh > state[0]): # 5 point local maximum
				bif[0].append(inp)
				bif[1].append(xhh)
	return bif


def generate_signal(ders, n, sampling, number_of_variables=1, warmup_time=1000.0, tau=3.0, eps=0.5, dt=0.01):
	"""generates signal for the oscillator driven by correlated noise
	
	:param ders: a list of state variable derivatives
	:param n: length of time series
	:param sampling: the sampling rate
	:param number_of_variables: the number of variables returned, not including the input (default 1)
	:param warmup_time: the time for relaxing to the stationary regime (default 1000)
	:param tau: noise correlation time (default 3.0)
	:param eps: noise strength (default 0.5)
	:param dt: time step (default 0.01)
	:returns: time series of the signal and driving noise"""
	# initial conditions
	state = [random.gauss(0,0.2) for i in range(len(ders))]
	state[0] = state[0] + 1.0
	resS = [[] for i in range(len(ders)+1)] # +1 for driving signal
	I = 0.0
	# warmup
	for i in range(round(warmup_time/dt)):
		I = I - (I/tau - eps*sqrt(2/tau)*random.gauss(0,1)/sqrt(dt))*dt
		state = one_step_integrator(state, ders, I, dt)
	# real integration
	for i in range(n*sampling):
		I = I - (I/tau - eps*sqrt(2/tau)*random.gauss(0,1)/sqrt(dt))*dt
		state = one_step_integrator(state, ders, I, dt)
		for c in range(len(ders)):
			resS[c].append(state[c])
		resS[-1].append(I)
	# return list
	rl = [resS[i][::sampling] for i in range(number_of_variables)]
	rl.append(resS[-1][::sampling])
	return rl


def generate_signal_e(ders, n, sampling, number_of_variables=1, warmup_time=1000.0, tau=3.0, eps=0.5, e_eps=1.0, dt=0.01):
	"""generates signal for the oscillator driven by e^(correlated noise)
	
	:param ders: a list of state variable derivatives
	:param n: length of time series
	:param sampling: the sampling rate
	:param number_of_variables: the number of variables returned, not including the input (default 1)
	:param warmup_time: the time for relaxing to the stationary regime (default 1000)
	:param tau: noise correlation time (default 3.0)
	:param eps: noise strength (default 0.5)
	:param e_eps: noise e strenght (default 1.0)
	:param dt: time step (default 0.01)
	:returns: time series of the signal and driving noise"""
	# initial conditions
	state = [random.gauss(0,0.2) for i in range(len(ders))]
	state[0] = state[0] + 1.0
	resS = [[] for i in range(len(ders)+1)] # +1 for driving signal
	It = 0.0
	# warmup
	for i in range(round(warmup_time/dt)):
		It = It - (It/tau - eps*sqrt(2/tau)*random.gauss(0,1)/sqrt(dt))*dt
		I = e_eps*exp(It)
		state = one_step_integrator(state, ders, I, dt)
	# real integration
	for i in range(n*sampling):
		It = It - (It/tau - eps*sqrt(2/tau)*random.gauss(0,1)/sqrt(dt))*dt
		I = e_eps*exp(It)
		state = one_step_integrator(state, ders, I, dt)
		for c in range(len(ders)):
			resS[c].append(state[c])
		resS[-1].append(I)
	# return list
	rl = [resS[i][::sampling] for i in range(number_of_variables)]
	rl.append(resS[-1][::sampling])
	return rl



from setuptools import setup

setup(
	name='oscillator_snap',
	packages=['oscillator_snap'],
	version='1.0',
	description='Make a snapshot of an oscillator through its time series',
	long_description=open('README.rst').read(),
	author='Rok Cestnik',
	author_email='rokcestn@uni-potsdam.de',
	license='MIT',
	install_requires=[
		'tensorflow',
		'keras',
		'numpy',
		'scipy',
		'h5py',
		'dill'
	],
)



import os
import dill
import numpy as np
import copy
from math import pi, log, exp, sqrt, floor
from keras.models import Sequential
from keras.layers import LSTM, GRU, Dense
from keras import optimizers

from oscillator_snap.oscillator_auxiliaries import *


def compile_model(model, learning_rate):
	"""compiles the model
	
	:param model: RNN model
	:returns: compiled model"""
	# optimizer (stochastic gradient descent)
	sgd = optimizers.SGD(lr=learning_rate, momentum=0.0, decay=0.0, nesterov=False)
	# compile model
	model.compile(loss='mean_squared_error', optimizer=sgd, metrics=['accuracy'])
	return model


def generate_model(data_dim_in, data_dim_out, past, nodes, learning_rate, cell=LSTM, n_hidden_layers=1):
	"""generates the model with n hidden layers (does not compile it)
	
	:param n_hidden_layers: number of hidden layers
	:returns: return the keras model"""
	model = Sequential()
	if(n_hidden_layers > 0): 
		model.add(cell(nodes, activation='tanh', return_sequences=True, input_shape=(past, data_dim_in)))  # input layer
		for i in range(n_hidden_layers-1):
			model.add(cell(nodes, activation='tanh', return_sequences=True))  # hidden layers
		model.add(cell(nodes, activation='tanh'))  # last hidden layer
	else:
		model.add(cell(nodes, activation='tanh', input_shape=(past, data_dim_in)))  # input layer
	model.add(Dense(data_dim_out, activation='linear'))  # output layer
	return model


def parse_train_data(seq, past, dim_in, dim_out):
	"""parses the training data for the RNN
	
	:param seq: the sequence to be parsed
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:param dim_out: the dimensionality of the prediction
	:returns: input data X and prediction Y""" 
	# network input output
	X = []
	Y = []
	# determine the length of the sequence
	L = len(seq[0])
	# take care of dimensions
	seq_in = seq[0:dim_in]
	seq_out = seq[0:dim_out]
	# reshape
	seq_in = np.array(seq_in)
	seq_in = seq_in.transpose()
	seq_out = np.array(seq_out)
	seq_out = seq_out.transpose()
	# organize
	for i in range(L-past):
		X.append(seq_in[i:i+past])
		Y.append(seq_out[i+past])
	return np.array(X), np.array(Y)


def forecast_starter(seq, past, dim_in):
	"""prepares the staring vector for RNN forecast
	
	:param seq: the sequence from which the beggining is used as the forcast starter
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:returns: starter input data""" 
	# determine the length of the sequence
	L = len(seq[0])
	# take care of dimensions
	seq_in = seq[0:dim_in]
	# reshape
	seq_in = np.array(seq_in)
	seq_in = seq_in.transpose()
	return np.array(seq_in[0:past])


def update_x(model, x, new_input, number_of_variables=1):
	"""update x and make the next prediction y (auxiliary for forecast) 
	
	:param model: RNN model
	:param x: input vector to RNN
	:param new_input: value of the new input
	:param number_of_variables: the number of variables that have to be updated (default 1)
	:returns: updated x""" 
	y = (model.predict(x))[0][0:number_of_variables] # next prediction value
	# update x
	x = x[0][1:]
	x = np.append(x, np.append(y, new_input)).reshape(1, x.shape[0]+1, x.shape[1])
	return x, y


def forecast(model, past, dim_in, stream_start, time_span, oscillator_input, number_of_variables=1):
	"""forecasts the signal, using the RNN
	
	:param model: RNN model
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:param forcast_start: the initial input to the network
	:param time_span: the time span of the forcasting, in network steps
	:param oscillator_input: the time series of the input being fed to the oscillator while forcasting (starting from the time of the first forcasted value)
	:param number_of_variables: the number of variables that are being forecasted (default 1)
	:returns: forcast of the time series""" 
	print("forecast")
	s_for = []
	x = np.array([stream_start])
	for i in range(time_span):
		x, y = update_x(model, x, oscillator_input[i], number_of_variables=number_of_variables)
		s_for.append(y.tolist()) # save the values
	return s_for


def period_measure(model, past, dim_in, forcast_start, constant_input_offset, thr=0.0, period_counts=100):
	"""estimate the natural period from the RNN model, also useful for chaotic and quasi-periodic dynamics because of averaging (in units of dt*sampling)
	
	:param model: RNN model
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:param forcast_start: the starting point of the time series
	:param constant_input_offset: the constant input to be fed to the network
	:param thr: signal threshold (default 0.0)
	:param period_counts: how many periods do we average over (default 100)
	:returns: the natural period""" 
	print("period estimation")
	x = np.array([forcast_start])
	yh, y = 0, 0 # initial setting of signal and its previous value
	# first warmup
	for t in range(1000):
		yh = y # previous value
		x, y = update_x(model, x, constant_input_offset) # update x,y
	# then we get to a position just after a crossing
	failsafe_time = 0
	while((yh < thr and y > thr) == False):
		yh = y # previous value
		x, y = update_x(model, x, constant_input_offset) # update x,y
		# time update
		failsafe_time = failsafe_time + 1
		# if it runs for too long it might be that it does not oscillate
		if(failsafe_time == 1000): # the choice 1000 is arbitrary
			print("\tallert: it will probably never reach the threshold")
			return
	previous_crossing = (thr-yh)/(y-yh)*1 # time of crossing correction
	# now set time to zero and look for following crossings
	time = 0
	avg_period = 0
	for p in range(period_counts):
		yh = thr+1 # break the condition
		failsafe_time = 0
		while((yh < thr and y > thr) == False):
			yh = y # previous value
			x, y = update_x(model, x, constant_input_offset) # update x,y
			# time update
			time = time + 1
			failsafe_time = failsafe_time + 1
			# if it runs for too long it might be that it does not oscillate
			if(failsafe_time == 1000): # the choice 1000 is arbitrary
				print("\tallert: it will probably never reach the threshold")
				return
		crossing = time + (thr-yh)/(y-yh)*1 # time of crossing
		avg_period = avg_period + crossing-previous_crossing # add to the period
		previous_crossing = crossing # reset previous crossing
	avg_period = avg_period/period_counts
	return avg_period


def PRC_measure(model, past, dim_in, forcast_start, constant_input_offset=0.0, warmup_time=1000, warmup_between_probes=100, stimulation=0.25, thr=0.0, period_counts=5, phase_repeats=20):
	"""estimate the PRC from the RNN model (in units of sampling)
	
	:param model: RNN model
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:param forcast_start: the starting point of the time series
	:param constant_input_offset: the constant input to be fed to the network (default 0.0)
	:param warmup_time: the time for the system to relax to its steady state (default 1000)
	:param warmup_between_probes: the time for the system to relax between diferent phase probes (default 100)
	:param stimulation: how strong to stimulate the oscillator (default 0.05)
	:param thr: signal threshold (default 0.0)
	:param period_counts: how many periods do we average over (default 5)
	:param phase_repeats: how many times to stimulate at each phase (default 10)
	:returns: the PRC in list format""" 
	print("PRC estimation")
	period = period_measure(model, past, dim_in, forcast_start, constant_input_offset, thr=thr)
	PRC = [[2*pi*(i+1)/period for i in range(floor(period)-1)],[0 for i in range(floor(period)-1)]] # PRC list (i+1 because when i=0 the hit comes effectively between 0.5 and 1.5 depending on how it crosses the threshold)
	x = np.array([forcast_start])
	yh, y = 0, 0 # initial setting of signal and its previous value
	# first warmup
	for t in range(warmup_time):
		yh = y # previous value
		x, y = update_x(model, x, constant_input_offset) # update x,y
	for ph in range(floor(period)-1):
		print("\tphase = ", ph, "/", floor(period)-1-1)
		# warmup between different phase probes
		for t in range(warmup_between_probes):
			yh = y # previous value
			x, y = update_x(model, x, constant_input_offset) # update x,y
		# set phase shift to 0, then we will add to it to get an average
		phase_shift = 0
		for r in range(phase_repeats):
			# then we get to a position just after a crossing
			while((yh < thr and y > thr) == False):
				yh = y # previous value
				x, y = update_x(model, x, constant_input_offset) # update x,y
			first_crossing = (thr-yh)/(y-yh)*1 # time of crossing correction
			# now set time to zero and look for following crossings
			time = 0
			# wait ph steps...
			for t in range(ph):
				yh = y # previous value
				x, y = update_x(model, x, constant_input_offset) # update x,y
				# time update
				time = time + 1
			# and then stimulate
			yh = y # previous value
			x, y = update_x(model, x, constant_input_offset+stimulation) # notice +stimulation
			# time update
			time = time + 1
			# now run for some periods and then evaluate the phase shift
			for p in range(period_counts):
				yh = thr+1 # break the condition (just in case - generally should be already broken)
				while((yh < thr and y > thr) == False):
					yh = y # previous value
					x, y = update_x(model, x, constant_input_offset) # update x,y
					# time update
					time = time + 1
				crossing = time + (thr-yh)/(y-yh)*1 # time of crossing
			# altered periods
			altered_periods = crossing-first_crossing
			# phase shift
			phase_shift = phase_shift + 2*pi*(altered_periods-period_counts*period)/period
		phase_shift = phase_shift/phase_repeats
		PRC[1][ph] = phase_shift/stimulation
	return PRC


def lyapunov_measure(model, past, dim_in, forcast_start, constant_input_offset=0.0, warmup_time=1000, delta=0.005, tau=10, unconsidered_trials=50, trials=500):
	"""estimate the largest Lyapunov exponent from the RNN model (in units of 1/(sampling*dt) ), 
	this is achieved by staring with two close trajectories (reference one - x, and perturbed one x_p), evolving them and evaluating the deviation. 
	To assure we are measuring the maximal exponent the perturbed trajectory is every time renormalized 
	(the whole past is rescaled by delta*sqrt(1/past*sum_i^past (x_p(t-i)-x(t-i))^2) ), 
	thus allowing the maximal exponent to take over and there's no need for any embedding - pretty neat:)
	
	:param model: RNN model
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:param forcast_start: the starting point of the time series
	:param constant_input_offset: the constant input to be fed to the network (default 0.0)
	:param warmup_time: the time for the system to relax to its steady state (default 1000)
	:param delta: the starting deviation (default 0.005)
	:param tau: the nuber of updates before the deviation is evaluated (default 10)
	:param uncosidered_trials: the number of trials that are not considered in the averaging, alowing the maximal Lyapunov exponent to take over (default 50)
	:param trials: the number of trials considered in the averaging (default 500)"""
	print("Lyapunov estimation")
	# warmup 
	x = np.array([forcast_start])
	for t in range(warmup_time):
		x, y = update_x(model, x, constant_input_offset) # update x,y
	# copy the trajectory
	x2 = x.copy()
	# first step, x2 gets perturbed slightly
	x, y = update_x(model, x, constant_input_offset) # update x,y
	x2, y2 = update_x(model, x2, constant_input_offset+delta) # update x2,y2, notice +delta
	for utr in range(unconsidered_trials): # unconsidered trials
		for t in range(tau):
			x, y = update_x(model, x, constant_input_offset) # update x,y
			x2, y2 = update_x(model, x2, constant_input_offset) # update x2,y2
		# rescale
		rescaling = sqrt(1/past*sum((x2[0][i][0]-x[0][i][0])**2 for i in range(len(x[0]))))
		x2 = x + delta/rescaling*(x2-x)
	lyapunov = 0
	for tr in range(trials): # considered trials	
		for t in range(tau):
			x, y = update_x(model, x, constant_input_offset) # update x,y
			x2, y2 = update_x(model, x2, constant_input_offset) # update x2,y2
		# rescale
		rescaling = sqrt(1/past*sum((x2[0][i][0]-x[0][i][0])**2 for i in range(len(x[0]))))
		x2 = x + delta/rescaling*(x2-x)
		# averaging deviation
		lyapunov += log(rescaling/delta)
	return lyapunov/tau/trials


def bifurcation_diagram(model, past, dim_in, forcast_start, ci_min, ci_max, dci, time_window=500):	
	"""measure the bifurcation diagram from the RNN model
	
	:param model: RNN model
	:param past: the number of considered steps in the past
	:param dim_in: the dimensionality of the data fed to the network
	:param forcast_start: the starting point of the time series
	:param ci_min: minimal (starting) value of the constant input - bifurcation parameter
	:param ci_max: maximal (ending) value of the constant input - bifurcation parameter
	:param dci: bifurcatin parameter step (ci - constant input)
	:param time_window: the time window in which the diagram is evaluated for each bifurcation parameter value, in network steps (default 500)
	:returns: the scatter plot of the bifurcation diagram""" 
	print("bifurcation diagram")
	bif = [[],[]] # bifurcation list
	x = np.array([forcast_start])
	# first warmup
	for t in range(1000):
		x, y = update_x(model, x, ci_min) # update x,y
	# bp loop (bifurcation parameter)
	for bp in [ci_min+dci*i for i in range(floor((ci_max-ci_min)/dci))]:
		print("\tbp = ", "%.3f" % bp, "/", ci_max)
		yhhhh = y # pre-pre-pre-previous value
		yhhh = y # pre-pre-previous value
		yhh = y # pre-previous value
		yh = y # previous value
		# warmup
		for t in range(100):
			x, y = update_x(model, x, bp) # update x,y
		for t in range(time_window):
			yhhhh = yhhh
			yhhh = yhh
			yhh = yh
			yh = y
			x, y = update_x(model, x, bp) # update x,y
			if(yhhhh < yhhh and yhhh < yhh and yhh > yh and yh > y): # 5 point local maximum
				bif[0].append(bp)
				bif[1].append(yhh)	
	return bif

	
def save_model_h5py(model, save_path="model_save"):
	"""saves a keras model to h5py file (also makes the 'model_save/' directory if it wasn't there before)
	
	:param model: model to save
	:param save_path: path where to save model, staring from the directory of the simulation (default "model_save")"""
	# directory
	if not os.path.exists(save_path):
		os.makedirs(save_path)
	# serialize model to JSON
	model_json = model.to_json()
	with open(save_path+"/model.json", "w") as json_file:
		json_file.write(model_json)
	# serialize weights to HDF5
	model.save_weights(save_path+"/model.h5")


def save_model_dill(model, save_path="model_save"):
	"""saves a keras model with dill (also makes the 'model_save/' directory if it wasn't there before)
	
	:param model: model to save
	:param save_path: path where to save model, staring from the directory of the simulation (default "model_save")"""
	# directory
	if not os.path.exists(save_path):
		os.makedirs(save_path)
	config = model.get_config()
	weights = model.get_weights()
	# pickling
	fc = open(save_path+'/config.pickle', 'wb')
	fw = open(save_path+'/weights.pickle', 'wb')
	dill.dump(config, fc)
	dill.dump(weights, fw)
	fc.close()
	fw.close()


def load_model_h5py(load_path="model_save"):
	"""loads model with h5py file (from 'model_save/' directory)
	
	:param load_path: path where to load model from, staring from the directory of the simulation (default "model_save")
	:returns: keras model"""
	# load json and create model
	json_file = open(load_path+'/model.json', 'r')
	model_json = json_file.read()
	json_file.close()
	model = model_from_json(model_json)
	# load weights into new model
	model.load_weights(load_path+"/model.h5")
	return model


def load_model_dill(load_path="model_save"):
	"""loads model with dill (from 'model_save/' directory)
	
	:param load_path: path where to load model from, staring from the directory of the simulation (default "model_save")
	:returns: keras model"""
	# unpickling
	fc = open(load_path+'/config.pickle', 'rb')
	fw = open(load_path+'/weights.pickle','rb')
	config = dill.load(fc)
	weights = dill.load(fw)
	fc.close()
	fw.close()
	# make model
	model = Sequential.from_config(config)
	model.set_weights(weights)
	return model
	
